import java.util.Vector; 

/* Copyright K Lano 2012
   package: Utilities
*/ 

class Maplet
{ Object source;
  Object dest;  /* Usually both are named */

  Maplet(Object src, Object dst)
  { source = src;
    dest = dst; 
  }

  public void print_maplet()
  { System.out.print(((Named) source).label);
    System.out.print("  |-->  ");
    System.out.println(((Named) dest).label); }

  public String toString()
  { return source.toString() + "  |-->   " + dest.toString(); } 

  public void display() 
  { System.out.print(source); 
    System.out.print("  |-->  ");
    System.out.println(dest); }

  public boolean equals(Object mm) 
  { if (mm instanceof Maplet) 
    { Maplet maplet = (Maplet) mm; 
      return (source.equals(maplet.source) && 
              dest.equals(maplet.dest)); } 
    else 
    { return false; } 
  } 
}

class Map
{
  int maxmap = 100;
  Vector elements;  /* vector of maplets */

  Map()  { elements = new Vector(maxmap); }

  public void clear()
  { elements = new Vector(maxmap); }

  public void add_element(Maplet m)
  {  elements.addElement(m);  /* assumes still a map */
  }

  public void add_pair(Object o1, Object o2)
  { Maplet mm = new Maplet(o1,o2);
    elements.addElement(mm); }

 public void print_map()
 { for (int i = 0; i < elements.size(); i++)
   { ((Maplet) elements.elementAt(i)).print_maplet(); }
 }

  public boolean in_domain(Object x)
  {  for (int i = 0; i < elements.size(); i++)
      { if (x == ((Maplet) elements.elementAt(i)).source)
         { return true; } }
      return false; 
  }

  public boolean in_range(Object x)
  {  for (int i = 0; i < elements.size(); i++)
      { if (x == ((Maplet) elements.elementAt(i)).dest)
         { return true; } }
      return false; 
  }

  public Vector domain()
  { Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++)
    { res.add(((Maplet) elements.elementAt(i)).source); }
    return res; 
  }
 
  public Vector range()
  { Vector res = new Vector(); 
    for (int i = 0; i < elements.size(); i++)
    { res.add(((Maplet) elements.elementAt(i)).dest); }
    return res; 
  }

  public Object apply(Object x)
  {  for (int i = 0; i < elements.size(); i++)
      { if (x == ((Maplet) elements.elementAt(i)).source)
         { return ((Maplet) elements.elementAt(i)).dest; } }
      return null; }

  public static Vector compose(Vector map1, Vector map2)
  { Vector res = new Vector();
    for (int i = 0; i < map1.size(); i++)
    { Maplet mm1 = (Maplet) map1.elementAt(i);
      for (int j = 0; j < map2.size(); j++)
      { Maplet mm2 = (Maplet) map2.elementAt(j);
        if (mm1.dest.equals(mm2.source))
        { Maplet mm3 = new Maplet(mm1.source, mm2.dest);
          res.add(mm3); } } }
    return res; }

  public void put(Object src, Object trg)
  { Map map2 = new Map(); 
    map2.add_pair(src,trg); 
    Vector newelems = compose(elements,map2.elements); 
    elements = VectorUtil.union(elements,newelems); 
  } 

  public static Vector transitiveClosure(Vector map)
  { Vector res = (Vector) map.clone();
     for (int i = 0; i < map.size(); i++)
     { Vector res1 = compose(res,map);
       Vector res2 = VectorUtil.vector_merge(res,res1);
       System.out.println("Trans closure step " + i + ": " + res2);
       res = res2; }
     return res; }   

}

