import java.util.Vector; 

/* Copyright K. Lano 2015
   Package: QVT
*/ 

public class PropertyTemplateItem
{ Attribute referredProperty; // may represent association
  Expression value = null;
  ObjectTemplateExp template = null;

  public PropertyTemplateItem(Attribute att)
  { referredProperty = att; }

  public Object clone()
  { PropertyTemplateItem res = new PropertyTemplateItem(referredProperty); 
    if (value != null) 
    { Expression newvalue = (Expression) value.clone(); 
      res.value = newvalue; 
    } 
    if (template != null) 
    { ObjectTemplateExp ote = (ObjectTemplateExp) template.clone(); 
      res.template = ote; 
    } 
    return res; 
  } 

  public void setValue(Expression v)
  { value = v; }

  public void setTemplate(ObjectTemplateExp t)
  { template = t; }

  public void setEntity(Entity e)
  { referredProperty.setEntity(e); }

  public Expression toSourceExpression(Vector bound, Expression contextObj)
  { // Expresses referredProperty = value, or 
    // referredProperty = template, as UML-RSDS expression
   Expression result = new BasicExpression("true");
   if (referredProperty == null) 
   { System.err.println("No feature defined for " + this); 
     return result; 
   } 

   BasicExpression fexp = new BasicExpression(contextObj, referredProperty);
   String op = "=";
   if (referredProperty.isMultiple())
   { op = ":"; }
    
   if (value != null)
   { if (value.isVariable()) 
     { // Expression valueexp = new BasicExpression(value);
       if (bound.contains(value + ""))
       { if ("=".equals(op))
         { result = new BinaryExpression("=", fexp, value);  }
         else 
         { result = new BinaryExpression(op, value, fexp); }
       }
       else
       { result = new BinaryExpression(op, value, fexp); 
         bound.add(value + ""); 
       }
    }
    else // just a value expression
    { result = new BinaryExpression("=", fexp, value);  }
  }
  else if (template != null)
  { Entity f = template.getEntity();
    Attribute x = template.bindsTo;
    Expression xexp = new BasicExpression(x);
    result = new BinaryExpression(op, xexp, fexp); 
    bound.add(x.getName());
    if (template.isEmpty()) { }
    else
    { Expression andcond = template.toSourceExpression(bound,xexp);
      result = new BinaryExpression("&", result, andcond); 
    }
  }
  return result;
 }

  public Expression toTargetExpression(Vector bound, Expression contextObj, Expression setting)
  { // Expresses referredProperty = value, or 
    // referredProperty = template, as UML-RSDS expression

   Expression result = new BasicExpression("true"); 
   if (referredProperty == null) 
   { System.err.println("No feature defined for " + this); 
     return result; 
   } 

   BasicExpression fexp = new BasicExpression(contextObj, referredProperty);
   String op = "=";
   if (referredProperty.isMultiple())
   { op = ":"; }
    
   if (value != null)
   { if (value.isVariable()) 
     { // Expression valueexp = new BasicExpression(value);
       if (bound.contains(value + ""))
       { if ("=".equals(op))
         { result = new BinaryExpression("=", fexp, value);  }
         else 
         { result = new BinaryExpression(op, value, fexp); }
       }
       else // should not occur
       { result = new BinaryExpression(op, value, fexp); 
         bound.add(value + ""); 
       }
    }
    else // just a value expression
    { result = new BinaryExpression("=", fexp, value);  }
  }
  else if (template != null)
  { // Entity f = template.getEntity();
    Attribute x = template.bindsTo;
    Expression xexp = new BasicExpression(x);

    // Expression setting;
    if (referredProperty.isMultiple())
    { setting = new BinaryExpression(":", xexp, fexp); }
    else
    { setting = new BinaryExpression("=", fexp, xexp); }

    if (template.isEmpty())
    { bound.add(x.getName());
      result = new BinaryExpression("&", result, setting); 
    }
    else
    { Expression andcond = template.toTargetExpression(bound,xexp,setting);
      result = new BinaryExpression("&", result, andcond);
    } 
  }
  return result;
 }

  public Expression toTargetExpressionOp(Vector bound, Expression contextObj, Expression setting)
  { // Expresses referredProperty = value, or 
    // referredProperty = template, as UML-RSDS expression

   Expression result = new BasicExpression("true"); 
   if (referredProperty == null) 
   { System.err.println("No feature defined for " + this); 
     return result; 
   } 

   BasicExpression fexp = new BasicExpression(contextObj, referredProperty);
   String op = "=";
   if (referredProperty.isMultiple())
   { op = ":"; }
    
   if (value != null)
   { if (value.isVariable()) 
     { // Expression valueexp = new BasicExpression(value);
       if (bound.contains(value + ""))
       { if ("=".equals(op))
         { result = new BinaryExpression("=", fexp, value);  }
         else 
         { result = new BinaryExpression(op, value, fexp); }
       }
       else // should not occur
       { result = new BinaryExpression(op, value, fexp); 
         bound.add(value + ""); 
       }
    }
    else // just a value expression
    { result = new BinaryExpression("=", fexp, value);  }
  }
  else if (template != null)
  { // Entity f = template.getEntity();
    Attribute x = template.bindsTo;
    Expression xexp = new BasicExpression(x);

    // Expression setting;
    if (referredProperty.isMultiple())
    { setting = new BinaryExpression(":", xexp, fexp); }
    else
    { setting = new BinaryExpression("=", fexp, xexp); }

    // bound.add(x.getName());
    if (template.isEmpty())
    { bound.add(x.getName());
      result = new BinaryExpression("&", result, setting); 
    }
    else
    { Expression andcond = template.toTargetExpressionOp(bound,xexp,setting);
      result = new BinaryExpression("&", result, andcond);
    } 
  }
  return result;
 }

 public boolean typeCheck(Vector entities, Vector types, Vector contexts, Vector env)
 { if (value != null) 
   { value.typeCheck(entities,types,contexts,env); } 
   if (template != null)
   { template.typeCheck(entities,types,contexts,env); } 
   return true; 
 } 
 
  public String toString()
  { String res = referredProperty + " = ";
    if (value != null)
    { res = res + value; }
    else // template != null
    { res = res + template; }
    return res; 
  }
}