import java.util.Vector; 

/* Copyright K. Lano 2015
   Package: QVT
*/ 

public class Relation extends QVTRule
{ boolean isTopLevel = true;
  Pattern when = null;
  Pattern where = null;
  Vector variable = new Vector(); // of Attribute

  // derived features: 
  Entity traceEntity = null;
  BehaviouralFeature nontopOp = null;


  public Relation(String nme, RelationalTransformation t)
  { super(nme, t); }

  public Object clone()
  { Relation res = new Relation(name, transformation); 
    res.isTopLevel = isTopLevel; 
    if (when != null) 
    { res.when = (Pattern) when.clone(); }  
    if (where != null) 
    { res.where = (Pattern) where.clone(); } 
    
    for (int k = 0; k < domain.size(); k++) 
    { RelationDomain rd = (RelationDomain) domain.get(k); 
      RelationDomain rd1 = (RelationDomain) rd.clone(); 
      rd1.rule = res; 
      res.addDomain(rd1); 
    } 
    
    return res; 
  } 

  public void addVariable(Attribute att)
  { variable.add(att); }

  public void setWhen(Pattern w)
  { when = w; }

  public void addWhen(Expression e)
  { when = new Pattern(); 
    when.addPredicate(e); 
  } // divide it up into a sequence of conjuncts

  public void setWhere(Pattern w)
  { where = w; }

  public void addWhere(Expression e)
  { where = new Pattern(); 
    where.addPredicate(e); 
  } // divide it up into a sequence of conjuncts

  public void setIsTop(boolean t)
  { isTopLevel = t; }

  public String toString()
  { String res = "relation";
    if (isTopLevel)
    { res = "top " + res; }
    res = res + " " + name + "\n";
    res = res + "{\n";
    for (int i = 0; i < variable.size(); i++)
    { Attribute att = (Attribute) variable.get(i);
      res = res + "  " + att.getName() + " : " + att.getType() + "\n";
    }
    for (int j = 0; j < domain.size(); j++)
    { Domain d = (Domain) domain.get(j);
      res = res + "  " + d + "\n";
    }
    if (when != null)
    { res = res + "  when { " + when + " }\n"; }
    if (where != null)
    { res = res + "  where { " + where + " }\n"; }
    res = res + "}\n\n";
    return res; 
  }

  public void addTraceEntity(Vector entities)
  { Entity e = getTraceEntity(); 
    if (e != null) 
    { entities.add(e); } 
  } 

  public Entity getTraceEntity()
  { String ename = getName() + "$trace";
    Entity res = new Entity(ename);
    for (int j = 0; j < domain.size(); j++)
    { RelationDomain d = (RelationDomain) domain.get(j);
      if (d.rootVariable != null)
      { res.addAttribute(d.rootVariable); }
    } // actually assocs. - addRole? 
    traceEntity = res;
    return res;
  }

  public BehaviouralFeature getNontopOp(UseCase transuc, Vector entities, Vector types)
  { if (isTopLevel)
    { return null; } 

    String opname = getName().toLowerCase() + "$op";
    Vector bound = new Vector();
    Vector pars = new Vector();
    for (int j = 0; j < domain.size(); j++)
    { RelationDomain d = (RelationDomain) domain.get(j);
      if (d.rootVariable != null)
      { pars.add(d.rootVariable); 
        bound.add(d.rootVariable.getName());
      }
    } 
    Expression ante = toSourceExpressionOp(bound);
    Expression succ = toTargetExpressionOp(bound);
    Expression post = new BinaryExpression("=>", ante, succ);
    BehaviouralFeature bf = new BehaviouralFeature(opname);
    // bf.setUseCase(transuc);
    bf.setParameters(pars); 
    bf.setQuery(false); 
    // bf.setEntity(traceEntity);
    // bf.setPost(post);
    bf.setPre(new BasicExpression("true"));
    bf.setPost(new BasicExpression("true"));
    nontopOp = bf;
    transuc.addOperation(bf);  
    bf.setUseCase(transuc);
    Constraint opact = new Constraint(ante,succ); 
    opact.typeCheck(types,entities,new Vector()); 
    opact.setUseCase(transuc); 
    opact.analyseConstraint(new Vector());
    Statement stat = opact.mapToDesign0(transuc); 
    bf.setActivity(stat);   
    return bf;
  }

  public boolean typeCheck(Vector types, Vector entities, Vector contexts, Vector env)
  { for (int i = 0; i < domain.size(); i++) 
    { Domain d = (Domain) domain.get(i); 
      d.typeCheck(types,entities,contexts,env); 
    }
    return true; 
  }  
   
  public Expression toSourceExpression(Vector bound, Vector entities)
  { // for top-level rules only
    Expression res = new BasicExpression("true");
    for (int i = 0; i < domain.size(); i++)
    { Domain d = (Domain) domain.get(i);
      if (d.isCheckable)
      { Expression ex = d.toSourceExpression(bound); 
        res = Expression.simplify("&", res, ex, null);
      }
    } // and when.whenToExp(bound, transformationRule)
    if (when != null) 
    { Expression wnexp = when.whenToExp(bound, transformation.rule, entities); 
      res = Expression.simplify("&", res, wnexp, null);
    } // and where as well 
    return res;
  }

  public Expression toInverseSuccedent(Vector bound, Vector entities)
  { // for top-level rules only
    Expression res = new BasicExpression("true");
    for (int i = 0; i < domain.size(); i++)
    { Domain d = (Domain) domain.get(i);
      if (d.isCheckable)
      { Expression ex = d.toTargetExpression(bound); 
        res = Expression.simplify("&", res, ex, null);
      }
    } 
    if (when != null) 
    { Expression wnexp = when.whenToExp(bound, transformation.rule, entities); 
      res = Expression.simplify("#&", res, wnexp, null);
    } // and where as well 
    if (where != null) 
    { Expression wexp = where.whereToExp(bound, transformation.rule); 
      res = Expression.simplify("#&", res, wexp, null);
    } 
    return res;
  }

  public Expression toTargetExpression(Vector bound)
  { // for top-level rules only
    Expression res = new BasicExpression("true");
    for (int i = 0; i < domain.size(); i++)
    { Domain d = (Domain) domain.get(i);
      if (d.isEnforceable)
      { Expression ex = d.toTargetExpression(bound); 
        res = Expression.simplify("#&", res, ex, null);
      }
    }
    if (where != null) 
    { Expression wexp = where.whereToExp(bound, transformation.rule); 
      res = Expression.simplify("#&", res, wexp, null);
    } 
    return res;
  } // add the # quantifiers at the outer level. 

  public Expression toInverseAntecedent(Vector bound)
  { // for top-level rules only
    Expression res = new BasicExpression("true");
    for (int i = 0; i < domain.size(); i++)
    { Domain d = (Domain) domain.get(i);
      if (d.isEnforceable)
      { Expression ex = d.toSourceExpression(bound); 
        res = Expression.simplify("&", res, ex, null);
      }
    }
    return res;
  }

  public Expression toSourceExpressionOp(Vector bound)
  { // for non top-level rules only
    Expression res = new BasicExpression("true");
    for (int i = 0; i < domain.size(); i++)
    { Domain d = (Domain) domain.get(i);
      if (d.isCheckable)
      { Expression ex = d.toSourceExpressionOp(bound); 
        res = Expression.simplify("&", res, ex, null);
      }
    } // and when.whenToExp(bound, transformationRule)
    return res;
  }

  public Expression toTargetExpressionOp(Vector bound)
  { // for non top-level rules only 
    Expression res = new BasicExpression("true");
    for (int i = 0; i < domain.size(); i++)
    { Domain d = (Domain) domain.get(i);
      if (d.isEnforceable)
      { Expression ex = d.toTargetExpressionOp(bound); 
        res = Expression.simplify("#&", res, ex, null);
      }
    }
    if (where != null) 
    { Expression wexp = where.whereToExp(bound, transformation.rule); 
      res = Expression.simplify("#&", res, wexp, null);
    } 
    return res;
  }

  public Constraint toConstraint(Vector entities, Vector types)
  { Vector bound = new Vector();
    if (isTopLevel)
    { Expression ante = toSourceExpression(bound,entities);
      Expression succ = toTargetExpression(bound);
      Expression fconj = ante.firstConjunct(); 
      Expression rem = ante.removeFirstConjunct(); 
      // System.out.println(fconj); 
      if (fconj != null && (fconj instanceof BinaryExpression))
      { BinaryExpression fcnj = (BinaryExpression) fconj; 
        if (":".equals(fcnj.operator) && fcnj.left instanceof BasicExpression)
        { Expression newante = rem.dereference((BasicExpression) fcnj.left);
          Expression newsucc = succ.dereference((BasicExpression) fcnj.left);
          Constraint res = new Constraint(newante, newsucc); 
          res.setOwner(fcnj.right.entity); 
          return res; 
        } 
      } 
      Constraint con = new Constraint(ante,succ);
      return con;
    }
    return null; // make an operation
  }

  public Expression deleteEnforceableDomainsExp()
  { Expression res = new BasicExpression("true"); 
    for (int i = 0; i < domain.size(); i++) 
    { RelationDomain rd = (RelationDomain) domain.get(i); 
      if (rd.isEnforceable)
      { Expression delrd = rd.deleteDomainExp(); 
        res = Expression.simplify("&",res,delrd,null); 
      } 
    } 
    return res; 
  } 
      
  public Constraint toInverseConstraint(Vector entities, Vector types)
  { Vector bound = new Vector();
    if (isTopLevel)
    { Expression ante = toInverseAntecedent(bound);
      Expression succ = toInverseSuccedent(bound,entities);
      Expression fconj = ante.firstConjunct(); 
      Expression rem = ante.removeFirstConjunct(); 
      // System.out.println(fconj); 
      if (fconj != null && (fconj instanceof BinaryExpression))
      { BinaryExpression fcnj = (BinaryExpression) fconj; 
        if (":".equals(fcnj.operator) && fcnj.left instanceof BasicExpression)
        { Expression newante = rem.dereference((BasicExpression) fcnj.left);
          Expression newsucc = succ.dereference((BasicExpression) fcnj.left);
          Expression notsucc = new UnaryExpression("not", newsucc); 
          Expression deldoms = deleteEnforceableDomainsExp(); 
          Expression invsucc = deldoms.dereference((BasicExpression) fcnj.left);
          Expression invante = Expression.simplify("&", newante, notsucc, null); 
          Constraint res = new Constraint(invante,invsucc);
          res.setOwner(fcnj.right.entity); 
          return res; 
        } 
      } 
      Expression notsucc = new UnaryExpression("not", succ); 
      Expression deldoms = deleteEnforceableDomainsExp(); 
      Expression invante = Expression.simplify("&", ante, notsucc, null); 
      Constraint con = new Constraint(invante,deldoms);
      return con;
    }
    return null; // make an operation
  }
}
