#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#define ALLOCATIONSIZE 10

#define TRUE 1
#define FALSE 0

/* This header provides ANSI C implementations of 
   OCL library operations. OCL indexing conventions 
   are used to access elements of collections and 
   strings: numbering starts from 1 instead of 0. 

   Note that all parameters may be regarded as const
   except in the case of treesort. 
*/


int length(void* col[])
{ int result = 0;
  if (col == NULL)  { return result; } 
  void* ex = col[0];
  while (ex != NULL)
  { result++;
    ex = col[result];
  }
  return result;
}

int isIn(void* x, void* col[])
{ int result = FALSE;
  if (col == NULL) { return result; } 
  void* ex = col[0];
  int i = 0; 
  while (ex != NULL)
  { if (ex == x) { return TRUE; } 
    i++; 
    ex = col[i];
  }
  return result;
}

void* any(void* col[])
{ int n = length(col); 
  if (n == 0) { return NULL; } 
  return col[0]; 
} 

void* at(void* col[], int i)
{ int n = length(col); 
  if (n == 0 || i > n) { return NULL; } 
  return col[i-1]; 
} 

void* first(void* col[])
{ int n = length(col); 
  if (n == 0) { return NULL; } 
  return col[0]; 
} 

void* last(void* col[])
{ int n = length(col); 
  if (n == 0) { return NULL; } 
  return col[n-1]; 
} 

char* concatenateStrings(const char* st1, const char* st2)
{ int n = strlen(st1);
  int m = strlen(st2);
  char* result = (char*) calloc(n + m + 1, sizeof(char));
  strcat(strcat(result, st1), st2);
  result[n+m] = '\0';
  return result;
}

unsigned char containsAll(void* col1[], void* col2[])
{ /* col2 <: col1 */
  int n2 = length(col2);
  unsigned char result = TRUE; 
  int i = 0;  
  for ( ; i < n2; i++)
  { void* ex = col2[i];
    if (ex == NULL) { return result; }
    if (isIn(ex,col1)) { }
    else { return FALSE; }
  }
  return result;
}

 unsigned char equalsSet(void* col1[], void* col2[])
 { return containsAll(col1,col2) && containsAll(col2,col1); }

 unsigned char equalsSequence(void* col1[], void* col2[])
 { unsigned char result = TRUE; 
   int n1 = length(col1);
   int n2 = length(col2);
   if (n1 != n2) { return FALSE; }
   int i = 0;  
   for ( ; i < n1; i++)
   { void* ex = col1[i];
     if (ex == NULL) { return result; }
     if (ex == col2[i]) { }
     else { return FALSE; }
   }
   return result;
 }

 unsigned char disjoint(void* col1[], void* col2[])
 { int n = length((void**) col1);
   unsigned char result = TRUE; 
   int i = 0;  
   for ( ; i < n; i++)
   { void* ex = col1[i];
     if (ex == NULL) { return result; }
     if (isIn(ex,col2)) { return FALSE; }
   }
   return result;
 }

int roundOcl(double x)
{ int f = (int) floor(x);
  int c = (int) ceil(x);
  if (x >= f + 0.5) { return c; }
  return f;
}

int floorOcl(double x)
{ return (int) floor(x); }

int ceilOcl(double x)
{ return (int) ceil(x); }

double cbrt(double x)
{ return pow(x, 1.0/3); }

char* toLowerCase(const char* s)
{ int n = strlen(s);
  char* result = (char*) calloc(n+1,sizeof(char));
  int i = 0;
  for ( ; i < n; i++)
  { result[i] = tolower(s[i]); }
  result[n] = '\0';
  return result;
}

char* toUpperCase(const char* s)
{ int n = strlen(s);
  char* result = (char*) calloc(n+1,sizeof(char));
  int i = 0;
  for ( ; i < n; i++)
  { result[i] = toupper(s[i]); }
  result[n] = '\0';
  return result;
}

char* subString(const char* s, int i, int j)
{ /* Characters of s from s[i-1] to s[j-1], inclusive */
  if (i > j) { return ""; }
  int n = strlen(s);
  char* result = (char*) calloc(j-i+2,sizeof(char));
    
  int x = i-1;
  for ( ; x < n && x < j; x++)
  { result[x-i+1] = s[x]; }
  result[j-i+1] = '\0';
  return result;
}

char* firstString(const char* s)
{ return subString(s,1,1); } 

char* lastString(const char* s)
{ int n = strlen(s); 
  return subString(s,n,n); 
} 

char* frontString(const char* s)
{ int n = strlen(s); 
  return subString(s,1,n-1);
} 

char* tailString(const char* s)
{ int n = strlen(s); 
  return subString(s,2,n); 
} 


  unsigned char startsWith(const char* s, const char* sb)
  { int n = strlen(s);
    int m = strlen(sb);
    if (m <= n && strncmp(s,sb,m) == 0)
    { return TRUE; }
    return FALSE;
  }

  unsigned char endsWith(const char* s, const char* sb)
  { int n = strlen(s);
    int m = strlen(sb);
    if (m <= n && strcmp(s + (n-m),sb) == 0)
    { return TRUE; }
    return FALSE;
  }

  int indexOfString(const char* s, const char* sb)
  { int n = strlen(s);
    int m = strlen(sb);
    int result = 0;  
    if (n == 0 || m == 0) { return result; }
    int i = 0;
    for ( ; i < n; i++)
    { if (startsWith(&s[i],sb))
      { return i+1; }
    }
    return result;
  }

  int count(void* x, void* col[])
  { int n = length(col);
    int result = 0; 
    int i = 0; 
    for ( ; i < n; i++)
    { if (x == col[i]) { result++; } }
    return result;
  }

  int countString(char* x, char* col[])
  { int n = length((void**) col);
    int result = 0; 
    int i = 0; 
    for ( ; i < n; i++)
    { char* y = col[i]; 
      if (strcmp(x,y) == 0) { result++; } 
    }
    return result;
  }

  char* reverseString(const char* s)
  { int n = strlen(s);
    char* result = (char*) calloc(n+1,sizeof(char));
    int x = n-1;
    int i = 0;
    for ( ; i < n; i++, x--)
    { result[i] = s[x]; }
    return result;
  }


  char** characters(char* s)
  { int n = strlen(s);
    char** result = (char**) calloc(n+1, sizeof(char*));
    int i = 0;
    for ( ; i < n; i++)
    { char* x = (char*) calloc(2, sizeof(char));
      x[0] = s[i];
      x[1] = '\0';
      result[i] = x;
    }
    result[n] = NULL;
	return result; 
  }

  char* subtractString(const char* s, const char* s1)
  { /* Characters of s not in s1 */
    int n = strlen(s);
    char* result = (char*) calloc(n+1,sizeof(char));
    int i = 0;
    int j = 0; 
    for ( ; i < n; i++)
    { char c = s[i];
      if (strchr(s1,c) == NULL) 
      { result[j] = c;
        j++;
      }
    }
    result[j] = '\0'; 
    return result;
  }


  int indexOf(void* x, void** col)
  { int n = length(col);
    int result = 0;  
    int i = 0;
    for ( ; i < n; i++)
    { if (col[i] == x)
      { return i+1; }
    }
    return result;
  }

   void** append(void* col[], void* ex)
   { void** result;
     int len = length(col);
     if (len % ALLOCATIONSIZE == 0)
     { result = (void**) calloc(len + ALLOCATIONSIZE + 1, sizeof(void*));
       int i = 0;
       for ( ; i < len; i++)
       { result[i] = col[i]; }
     } 
    else
    { result = col; }
    result[len] = ex;
    result[len+1] = NULL; 
    return result;
  }

  char** appendString(char* col[], char* ex)
  { char** result;
    int len = length((void**) col);
    if (len % ALLOCATIONSIZE == 0)
    { result = (char**) calloc(len + ALLOCATIONSIZE + 1, sizeof(char*));
      int i = 0;
      for ( ; i < len; i++)       
      { result[i] = col[i]; }
    } 
    else
    { result = col; }
    result[len] = ex;
    result[len+1] = NULL; 
    return result;
  }

  char** insertString(char* col[], char* ex)
  { char** result;
    if (isIn((void*) ex, (void**) col)) { return col; } 

    int len = length((void**) col);
    if (len % ALLOCATIONSIZE == 0)
    { result = (char**) calloc(len + ALLOCATIONSIZE + 1, sizeof(char*));
      int i = 0;
      for ( ; i < len; i++)       
      { result[i] = col[i]; }
    } 
    else
    { result = col; }
    result[len] = ex;
    result[len+1] = NULL; 
    return result;
  }

  char** newStringList(void)
  { char** result;
    result = (char**) calloc(ALLOCATIONSIZE + 1, sizeof(char*));
    result[0] = NULL; 
    return result;
  }

int* intSubrange(int a, int b)
{ if (b < a) { return NULL; }
  int n = b-a+1; 
  int* result = (int*) calloc(n, sizeof(int));
  int i = 0;
  for ( ; i < n; i++)
  { result[i] = a+i; }
  return result;
}

  int sumint(int col[], int n)
  { int result = 0;
    int i = 0;
    for ( ; i < n; i++)
    { result += col[i]; }
    return result;
  }

  int prdint(int col[], int n)
  { int result = 1;
    int i = 0;
    for ( ; i < n; i++)
    { result *= col[i]; }
    return result;
  }

  long sumlong(long col[], int n)
  { long result = 0;
    int i = 0;
    for ( ; i < n; i++)
    { result += col[i]; }
    return result;
  }

  long prdlong(long col[], int n)
  { long result = 1;
    int i = 0;
    for ( ; i < n; i++)
    { result *= col[i]; }
    return result;
  }

  double sumdouble(double col[], int n)
  { double result = 0;
    int i = 0;
    for ( ; i < n; i++)
    { result += col[i]; }
    return result;
  }

  double prddouble(double col[], int n)
  { double result = 1;
    int i = 0;
    for ( ; i < n; i++)
    { result *= col[i]; }
    return result;
  }

  int sumInt(int a, int b, int (*fe)(int))
   { int result = 0;
     int i = a;
    for ( ; i <= b; i++)
    { result += (*fe)(i); }
    return result;
  }

  int prdInt(int a, int b, int (*fe)(int))
   { int result = 1;
     int i = a;
     for ( ; i <= b; i++)
     { result *= (*fe)(i); }
     return result;
   }

  long sumLong(int a, int b, long (*fe)(int))
   { long result = 0;
     int i = a;
    for ( ; i <= b; i++)
    { result += (*fe)(i); }
    return result;
  }

  long prdLong(int a, int b, long (*fe)(int))
   { long result = 1;
     int i = a;
     for ( ; i <= b; i++)
     { result *= (*fe)(i); }
     return result;
   }

  double sumDouble(int a, int b, double (*fe)(int))
   { double result = 0;
     int i = a;
    for ( ; i <= b; i++)
    { result += (*fe)(i); }
    return result;
  }

  double prdDouble(int a, int b, double (*fe)(int))
   { double result = 1;
     int i = a;
     for ( ; i <= b; i++)
     { result *= (*fe)(i); }
     return result;
   }

  double maxdouble(double col[], int n)
   { int i = 1;
     double result = col[0];
     for ( ; i < n; i++)
     { if (col[i] > result)
       { result = col[i]; }
     }
     return result;
   }

  double mindouble(double col[], int n)
   { int i = 1;
     double result = col[0];
     for ( ; i < n; i++)
     { if (col[i] < result)
       { result = col[i]; }
     }
     return result;
   }

  int maxint(int col[], int n)
   { int i = 1;
     int result = col[0];
     for ( ; i < n; i++)
     { if (col[i] > result)
       { result = col[i]; }
     }
     return result;
   }

  int minint(int col[], int n)
   { int i = 1;
     int result = col[0];
     for ( ; i < n; i++)
     { if (col[i] < result)
       { result = col[i]; }
     }
     return result;
   }

  long maxlong(long col[], int n)
   { int i = 1;
     long result = col[0];
     for ( ; i < n; i++)
     { if (col[i] > result)
       { result = col[i]; }
     }
     return result;
   }

  long minlong(long col[], int n)
   { int i = 1;
     long result = col[0];
     for ( ; i < n; i++)
     { if (col[i] < result)
       { result = col[i]; }
     }
     return result;
   }

  char* sumString(char* col[])
  { int n = 0;
    int len = length((void**) col); 
    int i = 0;
    for ( ; i < len; i++)
    { n += strlen(col[i]); }
    char* result = (char*) calloc(n+1, sizeof(char));
    i = 0;
    for ( ; i < len; i++)
    { result = strcat(result, col[i]); }
    result[n] = '\0'; 
    return result;
  }

   char* maxString(char* col[])
   { int n = length((void**) col);
     if (n == 0) { return NULL; }
     char* result = col[0];
     int i = 0;
     for ( ; i < n; i++)
     { if (strcmp(result, col[i]) < 0)
       { result = col[i]; }
     }
     return result;
   }

   char* minString(char* col[])
   { int n = length((void**) col);
     if (n == 0) { return NULL; }
     char* result = col[0];
     int i = 0;
     for ( ; i < n; i++)
     { if (strcmp(result, col[i]) > 0)
       { result = col[i]; }
     }
     return result;
   }

  char* insertAtString(char* st1, int ind, char* st2)
  { if (ind <= 0) { return st1; }
    int n = strlen(st1);
    int m = strlen(st2);
    if (m == 0) { return st1; }

    char* result = (char*) calloc(n + m + 1, sizeof(char));
    int i = 0;
    for ( ; i < ind - 1 && i < n; i++)
    { result[i] = st1[i]; }
    if (i == ind - 1)
    { int j = 0; 
      for ( ; j < m; j++, i++)
      { result[i] = st2[j]; }
      for ( ; i < n + m; i++)
      { result[i] = st1[i - m]; }
    }
    else 
    { int j = 0; 
      for ( ; j < m; j++, i++)
      { result[i] = st2[j]; }
    }
    result[n+m] = '\0';
    return result;
  }

 int compareTo_String(const void* s1, const void* s2)
 { char* c1 = (char*) s1; 
   char* c2 = (char*) s2; 
   return strcmp(c1,c2); 
 } 

int compareTo_int(const void* s1, const void* s2) 
{ int* n1 = (int*) s1; 
  int* n2 = (int*) s2; 
  return *n2 - *n1; 
} 

int compareTo_long(const void* s1, const void* s2) 
{ long* n1 = (long*) s1; 
  long* n2 = (long*) s2; 
  if (*n1 < *n2) { return 1; } 
  if (*n1 == *n2) { return 0; } 
  return -1;  
} 

int compareTo_double(const void* s1, const void* s2) 
{ double* n1 = (double*) s1; 
  double* n2 = (double*) s2; 
  if (*n1 < *n2) { return 1; } 
  if (*n1 == *n2) { return 0; } 
  return -1;  
} 

int compareTo_boolean(const void* s1, const void* s2)
{ unsigned char* n1 = (unsigned char*) s1; 
  unsigned char* n2 = (unsigned char*) s2; 
  return (int) (*n2 - *n1); 
} 

struct ocltnode
{ void* object;
  struct ocltnode* left;
  struct ocltnode* right;
};

void printTree(struct ocltnode* self, char* (*tostr)(void*))
{ if (self != NULL)
  { printTree(self->left, tostr);
    printf("%s\n", (*tostr)(self->object));
    printTree(self->right, tostr);
  }
}

int copyTree(struct ocltnode* self, int i, void** res)
{ if (self != NULL) 
  { int j = copyTree(self->left, i, res); 
    res[j] = self->object; 
	j++; 
	return copyTree(self->right, j, res); 
  } 
  return i; 
}


struct ocltnode* addToTree(struct ocltnode* self, void* elem, int (*comp)(void*, void*))
{ if (self == NULL)
  { self = (struct ocltnode*) malloc(sizeof(struct ocltnode));
    self->object = elem;
    self->left = NULL;
    self->right = NULL;
  }
  else
  { int cond = (*comp)(elem, self->object);
    if (cond < 0)
    { self->left = addToTree(self->left, elem, comp); }
    else
    { self->right = addToTree(self->right, elem, comp); }
  }
  return self;
}

void** treesort(void* col[], int (*comp)(void*, void*))
{ if (col == NULL) { return NULL; }
  int len = length(col);
  void** res;
  res = (void**) calloc(len+1, sizeof(void*));
  struct ocltnode* tree = NULL;
  int i = 0;
  for ( ; i < len; i++)
  { tree = addToTree(tree, col[i], comp); }

  copyTree(tree, 0, res);
  res[len] = NULL; 
  return res;
}
 

void displayString(char* s)
{ printf("%s\n", s); }


void displayint(int s)
{ printf("%d\n", s); }

void displaylong(long s)
{ printf("%ld\n", s); }

void displaydouble(double s)
{ printf("%f\n", s); }

void displayboolean(unsigned int s)
{ if (s == TRUE) { printf("%s\n", "true"); } 
  else 
  { printf("%s\n", "false"); } 
} 
